----------------------------------------------------------------------------------
-- comtypes.adb: Paketimplementierung fr bentigte Typen und Funktion zur      --
--               Verwaltung eines Syntaxbaumes                                  --
--                                                                              --
-- Version 1.1                                                                  --
--  Feb / Mar 2008 by C. Rutetzki                                              --
---------------------------------------------------------------------------------- 

package body comtypes is
   -------------------------------------------------------------------------------
   -- TermType: Umwandeln vom Datentyp "TokenArt" in "Alphabet"
   
   function TermType (t : tokenart) return TTyp is
   begin
      case t is
         when ganzezahl_token => return zahlT;  
         when bezeichner_token => return nameT; 
         when rop_token => return ropT;
         when op_token => return opT;         
         when auf_token => return aufT;        
         when zu_token => return zuT;         
         when zuw_token => return zuweisungT;        
         when strichp_token => return semikolonT;    
         when if_token => return ifT;         
         when then_token => return thenT;       
         when else_token => return elseT;       
         when fi_token => return fiT;         
         when skip_token => return skipT;       
         when while_token => return whileT;      
         when do_token => return doT;         
         when od_token => return odT;
         -- Die Flle von true und false treten aufgrund fehlender Implementierung
         -- nicht auf (s. *.ads)
         when others => return falseT;
      end case;
      
   end TermType;
    
   --------------------------------------------------------------------------------
   -- Terminals: Modelliert Menge aller mglichen Terminale
   
   type Terminals is array (TTyp) of Boolean;

   --------------------------------------------------------------------------------
   -- prefTerms: Gibt zu einem bergebenen Nichtterminal die "Menge" aller 
   --            Terminale zurck, mit der eine aus dem Nichtterminal entstandene
   --            Terminalfolge beginnen kann.
   -- Bsp:       Ausdruck -> (aufT, zahlT, nameT)

   function prefTerms(nt : ntTyp) return Terminals is
      -- Rckgabearray
      terms : Terminals := (others => false);
   begin
      -- Terminale "abgrasen"
      case nt is
         when leereAnweisung =>
            terms(skipT) := true;
         when Zuweisung =>
            terms(nameT) := true;
         when bedingteAnweisung =>
            terms(ifT) := true;
         when Schleife =>
            terms(whileT) := true;
         when Ausdruck | Bedingung =>
            terms := (aufT | zahlT | nameT => true, others => false);
            
            -- False und True bei Bedingung
            if nt = Bedingung then
               terms(trueT) := true;
               terms(falseT) := true;
            end if;
                    
         when others =>
            terms := (ifT | nameT | skipT | whileT => true, others => false);
      end case;
      
      -- Rckgabe
      return terms;
   end prefTerms;
   
   --------------------------------------------------------------------------------
   -- createGB:  erstellt aus einer bergebenen Tokenliste eine Modellierung
   --            des entsprechenden Syntaxbaumes vom Datentyp GrBaum nach dem
   --            recursive-descendant-Parser-Prinzip. 
   --            Dazu existiert fr jedes Nichtterminal eine entsprechende Funktion
   --            die zunchst berprft, ob fr das Token an der gegenwrtigen 
   --            Leseposition (pPreview) das jeweilige Nichtterminal in Frage kommt.
   --            Kommt es in Frage wird der Baum erweitert und fallabhngig 
   --            die nchste "Nichtterminal"-Funktion aufgerufen oder im Falle
   --            einer Unstimmigkeit false zurckgegeben, was zwangslufig zu einem 
   --            leeren Baum (null) fhrt!
   -- Argumente: - tks:      Tokenliste (aus Scannerpak)
   --            - pTksLeft: bei Fehler enthlt pTksLeft die Rest-Token-Liste
   --            - error:    Fehlermeldung, sofern Fehler auftritt
   --            Die letzten beiden Argumente sind Zeiger, die standardmig mit
   --            null besetzt werden, wenn Fehlerangaben nicht bentigt werden.
   -- Bsp1:      Aus dem while-Programm 
   --            a := 2
   --            entstnde folgender Baum:
   --            gb : new GrKnoten'(Programm, null,
   --                  new GrKnoten'(AnwF, null,
   --                   new GrKnoten'(Anw, null,
   --                    new GrKnoten'(Zuweisung, null,
   --                     new GrKnoten'(nameT, 
   --                      new GrKnoten'(zuweisungT, 
   --                       new GrKnoten'(zahlT, null, "2"), ":="), "a")))));
   --            (Fr dieses Beispiel wurden Strings anstelle von 
   --            Unbounded_Strings fr "2", ":=" und "a" verwendet)
   -- Bsp2:      Aus dem while-Programm
   --            a := 2;
   --            entsteht ein leerer Baum (da Semikolon nur zwischen
   --            Anw und AnwF stehen darf). 
   --            (error: "Allgemeiner Parser Fehler: Kein vollstaendiges Programm!")
      
   function createGB(tks : listofToken;
                     pTksLeft :  PTOKENLIST := null;
                     error : PSTRING := null) return GrBaum is
      -- Rckgabebaum 
      pRtn : GrBaum := null;
      
      -- grundstzlich Zeiger auf das "unterste" Element
      pLast : GrBaum := null;
      
      -- Zeiger auf aktuelles Token (Vorschauelement)
      pPrev : listOfToken := tks;
   
      -- Position des Tokens pPrev (bei 1 beginnend)
      PrevPos : Natural := 1;

      -- Prozedur ermittelt nchstes Vorschauelement und erhht Zhler
      procedure NextPrev is
      begin
         -- nchstes Token
         if pPrev /= null then
            pPrev := pPrev.Rest;
            PrevPos := PrevPos + 1;
         end if;
      end NextPrev;
      
         
      
           
      -- CAusdruck bekanntmachen
      function CAusdruck (pLeft : GrBaum) return Boolean;

      ----------------------------------------------------------------------------
      -- CBedingung: prft ob Bedingung bei Vorschauelement mglich und ergnzt
      --             den Baum entsprechend. 
      -- Parameter:  pLeft: gibt an wo die Ergnzung stattfinden soll
      --                    Da Bedingung immer einen linken Nachbarn haben
      --                    muss (ifT oder whileT), wird Bedingung bei pLeft.rest 
      --                    angefgt.
      
      function CBedingung (pLeft : GrBaum) return Boolean is
         -- Mgliche Startterminale fr gegebenes Vorschauelement
         prefix : Terminals := prefTerms(Bedingung);
         
         -- Knoten an dem der Ausdruck angefgt werden soll
         pKn, pKnTmp : GrBaum := pLeft;
      begin
          -- leerer Baum?
         if pRtn = null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt 
         -- => Nichtterminal nchstes Baumelement
         if prefix(TermType(pPrev.token)) then
            
            -- neuen Knoten erstellen
            pkn.rest := new GrKnoten'(Bedingung, null, null);
            pkn := pkn.rest;
         else return false; 
         end if;
         
         -- Grammatikregel verwenden: 
         case TermType(pPrev.token) is
            when trueT =>
               pKn.knoten := new GrKnoten'(trueT, null, pPrev.string);
               NextPrev;--pPreview := pPreview.rest;
               return true;
            when falseT =>
               pKn.knoten := new GrKnoten'(falseT, null, pPrev.string);
               NextPrev;--pPreview := pPreview.rest;
               return true;
            when aufT | nameT | zahlT =>
               -- Dummy verwenden 
               pKnTmp := new GrKnoten'(Anw, null, Null);
               
               -- nchsten Ausdruck bestimmen
               if not CAusdruck(pKnTmp) or else pPrev = null or else
                     TermType(pPrev.token) /= ropT then
                                   
                  return false;
               else
                  -- Dummy nicht mehr bentigt
                  pKnTmp:= pKnTmp.rest;
                  
                  -- neuen Unterknoten einbinden
                  pKn.knoten := pKnTmp;
                  
               end if;
               
               -- Operation
               pKnTmp.rest := new GrKnoten'(ropT, null, pPrev.string);
               pKnTmp := pKnTmp.rest;
               
               -- Zeiger weitersetzen
               NextPrev;--pPreview := pPreview.rest;
               

               -- nchsten Ausdruck bestimmen
               if not CAusdruck(pKnTmp) then 
                  return false;
               else
                  pKnTmp := pKnTmp.rest;
               end if;
               
                             
               -- Rckgabe
               return true;
            when others =>
               return false;
            
            
         end case;
         
      end CBedingung;


      -- AnwF bekannt machen
      function CAnwF return Boolean;
      
      ----------------------------------------------------------------------------
      -- CSchleife: prft ob Schleife bei Vorschauelement mglich und ergnzt 
      --            den Baum entsprechend. 
      --            Dabei wird hier die Schleife an pLast.knoten angehngt
      
      function CSchleife return Boolean is
         -- Mgliche Startterminale fr gegebenes Vorschauelement
         prefix : Terminals := prefTerms(Schleife);
         
         -- Knoten an dem der Ausdruck angefgt werden soll
         pKn, pKnTmp : GrBaum := pLast;
         
         -- letztes Element zwischenspeichern
         pLastTmp : GrBaum := null;
      begin
          -- leerer Baum?
         if pRtn = null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt 
         -- => Nichtterminal nchstes Baumelement
         if prefix(TermType(pPrev.token)) then
    
            -- neuen Knoten erstellen
            pLast.knoten := new GrKnoten'(Schleife, null, null);
            pLast := pLast.knoten;
            pkn := pLast;
         
            -- whileT einbinden
            pKn.knoten := new GrKnoten'(whileT, null, pPrev.string);
            pKn := pKn.Knoten;
            
            -- auf die while-Anw zeigen
            pKnTmp := pKn;
                        
            -- Zeiger weitersetzen
            NextPrev;--pPreview := pPreview.rest;
               
            -- Bedingung bestimmen
            if not CBedingung(pKn) or else pPrev = null or else
                  TermType(pPrev.token) /= doT then
                     
               return false;
            else
              
               pKn := pKn.rest;
            end if;
               
            -- do
            pKn.rest := new GrKnoten'(doT, null, pPrev.string);
            pKn := pKn.rest;
               
            -- Zeiger weitersetzen
            NextPrev;--pPreview := pPreview.rest;
                     
            -- pLast sichern (da bei CAnwF pLast verndert wird)
            pLastTmp := pLast;   -- entspricht while-Anweisung
               
            -- pLast vorlufig ausklinken und Dummy erstellen, damit die AnwF
            -- spter an beliebiger Stelle platziert werden kann
            -- (pKnTmp.knoten zeigt am Ende auf neue AnwF)            
            pKnTmp := new GrKnoten'(Anw, null, null);
            pLast := pKnTmp;
            
            -- nchste AnwF bestimmen
            if not CAnwF or else pPrev = null or else
                  TermType(pPrev.token) /= odT then
                     
               return false;
              
            else
               -- pKn.rest auf neu erstellte Anweisungsfolge
               pKn.rest := pKnTmp.knoten;
               pKnTmp.knoten := null;
               

               -- pLast wieder auf whileT setzen
               pLast := pLastTmp;
               
               -- pKn auf AnwF
               pKn := pKn.rest;
               
               -- odT anfgen
               pKn.rest := new GrKnoten'(odT, null, pPrev.string);
               pKn := pKn.rest;
               
                         
            end if;
            
            -- Zeiger weiter setzen
            if pPrev /= null then
               NextPrev;--pPreview := pPreview.rest;
            end if;
            
      

            -- Rckgabe
            return true;
         else return false;
         
            
         end if;
         
      end CSchleife;


      ----------------------------------------------------------------------------
      -- CBedAnw: prft ob Bedingte Anweisung bei Vorschauelement mglich und  
      --          ergnzt den Baum entsprechend. 
      --          Dabei wird hier die Anweisung an pLast.knoten angehngt

      function CBedAnw return Boolean is
         -- Mgliche Startterminale fr gegebenes Vorschauelement
         prefix : Terminals := prefTerms(bedingteAnweisung);
         
         -- Knoten an dem der Ausdruck angefgt werden soll
         pKn, pKnTmp : GrBaum := pLast;
         
         -- letztes Element zwischenspeichern
         pLastTmp : GrBaum := null;
      begin
          -- leerer Baum?
         if pRtn = null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt
         -- => Nichtterminal nchstes Baumelement
         if prefix(TermType(pPrev.token)) then
    
            -- neuen Knoten erstellen
            pLast.knoten := new GrKnoten'(bedingteAnweisung, null, null);
            pLast := pLast.knoten;
            pkn := pLast;
         
            -- ifT einbinden
            pKn.knoten := new GrKnoten'(ifT, null, pPrev.string);
            pKn := pKn.Knoten;
            
            -- auf die if-Anw zeigen
            pKnTmp := pKn;
                        
            -- Zeiger weitersetzen
            NextPrev;--pPreview := pPreview.rest;
               
            -- Bedingung bestimmen
            if not CBedingung(pKn) or else pPrev = null or else
               TermType(pPrev.token) /= thenT then
               return false;
            else
              
               pKn := pKn.rest;
            end if;
               
            -- Then
            pKn.rest := new GrKnoten'(thenT, null, pPrev.string);
            pKn := pKn.rest;
               
            -- Zeiger weitersetzen
            NextPrev;--pPreview := pPreview.rest;
                     
            -- pLast sichern (zeigt auf ifT)
            pLastTmp := pLast;  
                           
            -- pLast vorlufig ausklinken und Dummy erstellen 
            -- (pKnTmp.knoten zeigt am Ende auf neue AnwF)
            pKnTmp := new GrKnoten'(Anw, null, null);  
            pLast := pKnTmp;
            
            -- nchsten AnwF bestimmen
            if not CAnwF or else pPrev = null or else
               TermType(pPrev.token) /= elseT then

               return false;
            else
               -- Anweisungsfolge an richtige Position verschieben
               
               -- pKN.rest ist nun auf der neu hinzugefgten AnwF
               pKn.rest := pKnTmp.knoten;
               pKnTmp.knoten := null;
               
               -- pLast wieder auf Dummy zeigen lassen
               pLast := pKnTmp;
               

               -- pKn auf AnwF
               pKn := pKn.rest;            -- auf AnwF
               
               
               -- pLast zeigt weiterhin auf Bedingung, pLastTmp auf if-Terminal

            end if;
            
           
               
            -- else
            pKn.rest := new GrKnoten'(elseT, null, pPrev.string);
            pKn := pKn.rest;
            NextPrev;--pPreview := pPreview.rest;
            
            
               
            -- nchsten AnwF bestimmen
            if not CAnwF or else pPrev = null or else
                  TermType(pPrev.token) /= fiT then
                
               return false;
            else
               -- Anweisungsfolge an richtige Position verschieben
               
               -- pKn.rest auf neu erstellte Anweisungsfolge
               pKn.rest := pKnTmp.knoten;
               pKnTmp.knoten := null;
               
               
              
               -- pLast wieder auf ifT setzen
               pLast := pLastTmp;
               
               -- pKn auf AnwF
               pKn := pKn.rest;
               
               -- fi anfgen
               pKn.rest := new GrKnoten'(fiT, null, pPrev.string);
               pKn := pKn.rest;
               
                         
            end if;
            
            -- Zeiger weiter setzen
            if pPrev /= null then
               NextPrev;--pPreview := pPreview.rest;
            end if;
            
      

            -- Rckgabe
            return true;
         else return false;
         
            
         end if;
         
      end CBedAnw;

      
      ----------------------------------------------------------------------------
      -- CAusdruck: prft ob Ausdruck bei Vorschauelement mglich und  
      --            ergnzt den Baum entsprechend. 
      -- Parameter: pLeft: Ausdruck wird an pLeft.rest angehngt
                         
      function CAusdruck (pLeft : GrBaum) return Boolean is
         -- Mgliche Startterminale fr gegebenes Vorschauelement
         prefix : Terminals := prefTerms(Ausdruck);
         
         -- Knoten an dem der Ausdruck angefgt werden soll
         pKn : GrBaum := pLeft;
      begin
          -- leerer Baum?
         if pRtn = null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt 
         -- => Nichtterminal nchstes Baumelement
         if prefix(TermType(pPrev.token)) then
            
            -- neuen Knoten erstellen
            pkn.rest := new GrKnoten'(Ausdruck, null, null);
            
            
            pkn := pkn.rest;
         else return false; 
         end if;
         
         -- Grammatikregel verwenden: 
         case TermType(pPrev.token) is
            when zahlT =>
               pKn.knoten := new GrKnoten'(zahlT, null, pPrev.string);
               NextPrev;--pPreview := pPreview.rest;
               
               return true;
            when nameT =>
               pKn.knoten := new GrKnoten'(nameT, null, pPrev.string);
               NextPrev;--pPreview := pPreview.rest;
               return true;
            when aufT =>
               -- Klammer einbinden
               pKn.knoten := new GrKnoten'(aufT, null, pPrev.string);
               pKn := pKn.Knoten;
               
               -- Zeiger weitersetzen
               if pPrev = null then return false;
               else NextPrev; end if;
               
               -- nchsten Ausdruck bestimmen
               if not CAusdruck(pKn) or else pPrev = null or else
                  TermType(pPrev.token) /= opT then
                  return false;
               else
                  pKn := pKn.rest;
               end if;
               
               -- Operation
               pKn.rest := new GrKnoten'(opT, null, pPrev.string);
               pKn := pKn.rest;
               
               -- Zeiger weitersetzen
               NextPrev;--pPreview := pPreview.rest;
               

               -- nchsten Ausdruck bestimmen
               if not CAusdruck(pKn) or else pPrev = null or else
                  TermType(pPrev.token) /= zuT then

                  return false;
               else
                  pKn := pKn.rest;
               end if;
               
               -- Klammer
               pKn.rest := new GrKnoten'(zuT, null, pPrev.string);
               pKn := pKn.rest;
               NextPrev;--pPreview := pPreview.rest;
   
               

               -- Rckgabe
               return true;
            when others => return false;
            
            
         end case;
         
         
         
      end CAusdruck;

     
      ----------------------------------------------------------------------------
      -- CZuweisung: prft ob Zuweisung bei Vorschauelement mglich und  
      --             ergnzt den Baum entsprechend. 
      
      function CZuweisung return Boolean is
                
      begin
          -- leerer Baum?
         if pRtn = null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt 
         -- => Nichtterminal nchstes Baumelement
         if TermType(pPrev.token) = nameT and then
            pPrev.rest /= null and then
               TermType(pPrev.rest.token) = zuweisungT and then
                  pPrev.rest.rest /= null then
                             
            -- neuen Knoten erstellen
            pLast.knoten := new GrKnoten'(Zuweisung, null,
                              new GrKnoten'(namet,
                  new GrKnoten'(zuweisungT, null, to_Unbounded_String(":=")),
                     pPrev.string));
            
            -- plast auf name zeigen lassen
            pLast := pLast.knoten.knoten;
 
            
            -- Zeiger zwei weitersetzen
            NextPrev; NextPrev;   --pPreview := pPreview.rest.rest;
            
            -- Ausdruck bestimmen
            return CAusdruck(pLast.rest);
         else return false;
         end if;
                  
                  
      end CZuweisung;
  
      ----------------------------------------------------------------------------
      -- CLeereAnw: prft ob leere Anweisung bei Vorschauelement mglich und  
      --            ergnzt den Baum entsprechend 

      function CLeereAnw return Boolean is
         -- Mgliche Startterminale fr gegebenes Vorschauelement
         prefix : Terminals := prefTerms(leereAnweisung);
      begin
         
          -- leerer Baum?
         if pRtn = null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt 
         -- => Nichtterminal nchstes Baumelement
         if prefix(TermType(pPrev.token)) then
            
            -- neuen Knoten erstellen
            pLast.knoten := new GrKnoten'(leereAnweisung, null, null);
            pLast := pLast.knoten;
            
            -- skip-Anweisung
            pLast.knoten := new GrKnoten'(skipT, null, pPrev.string);
            pLast := pLast.knoten;
            
            -- Zeiger weitersetzen
            NextPrev;--pPreview := pPreview.rest;
            
         else
            return false;            
         end if;
         
         
         
         -- Rckgabe
         return true;
         
                  
      end CLeereAnw;
      
      ----------------------------------------------------------------------------
      -- CAnw: prft ob Anw bei Vorschauelement mglich und  
      --       ergnzt den Baum entsprechend. 

      function CAnw return Boolean is
         -- Mgliche Startterminale fr gegebenes Vorschauelement
         prefix : Terminals := prefTerms(Anw);
      begin
         
          -- leerer Baum?
         if pRtn = null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt 
         -- => Nichtterminal nchstes Baumelement
         if prefix(TermType(pPrev.token)) then
            
            -- neuen Knoten erstellen
            pLast.knoten := new GrKnoten'(Anw, null, null);
            pLast := pLast.knoten;
         end if;
         
         -- Grammatikregel verwenden: 
         case TermType(pPrev.token) is
            when ifT =>
               return CBedAnw;
            when nameT =>
               return CZuweisung;
            when skipT =>
               return CLeereAnw;
            when whileT =>
               return CSchleife;
            when others => return false;
            
            
         end case;
         
      end CAnw;
      
      ----------------------------------------------------------------------------
      -- CAnwF: prft ob Anw bei Vorschauelement mglich und  
      --        ergnzt den Baum entsprechend. 
      
      function CAnwF return Boolean is
         -- Mgliche Startterminale fr gegebenes Vorschauelement
         prefix : Terminals := prefTerms(AnwF);
         
         -- Knoten merken
         pKn, pKnTmp : GrBaum := null;
         
         -- Token merken
         pTokenTmp : listofToken := null;
      begin
         
          -- leerer Baum?
         if pRtn = null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt 
         -- => Nichtterminal nchstes Baumelement
         if prefix(TermType(pPrev.token)) then
         

            -- neuen Knoten erstellen
            pLast.knoten := new GrKnoten'(AnwF, null, null);

            pLast := pLast.knoten;
            pKn := pLast; -- Knoten merken
         else return false;
            
         end if;
         
         -- temporren Knoten festhalten (Dort beginnt potentielle AnwF / Anw)
         pKnTmp := pLast;

         
         -- Grammatikregel verwenden
         if CAnw then                     

            -- Anweisung erfolgreich (vielleicht Anweisungsfolge)
            if pPrev /= null and then
               TermType(pPrev.token) = semikolonT then
               -- Token sichern
               pTokenTmp := pPrev;
               
               -- Semikolon aufnehmen (rechts von Anw)
               pKnTmp.knoten.rest := new GrKnoten'(semikolonT, null,
                  to_unbounded_String(";"));
               
               -- um das Semikolon weitersetzen
               NextPrev;--pPreview := pPreview.rest;
               
               -- Knoten festhalten (auf Semikolon links von pot AnwF)
               pKnTmp := pKnTmp.knoten.rest;
               
               -- plast beliebiges Nichtterminal (pLast gegenwrtig auf Terminal)
               pLast := new GrKnoten'(Anw, null, null);
               
               -- Position sichern
               pKn := pLast;
               
               -- Anweisungsfolge probieren
               if CAnwF then
                  
                  -- pLast zeigt nun auf letztes Terminal
                  -- pKn.knoten zeigt auf neue AnwF
                  -- pknTmp zeigt weiterhin auf Semikolon => pKnTmp.rest auf pKn.knoten
                  pknTmp.rest := pkn.Knoten;
                  
               else                  
                  -- Semikolon nach Anweisung ohne folge-Anweisung: Fehler
                  return false;
               end if;
            end if;
            
               
            -- Rckgabe
            return true;
         else
            return false;
         end if;     
         
      end CAnwF;
      
      ----------------------------------------------------------------------------
      -- CProgramm: prft ob Programm (Startsymbol) bei Vorschauelement mglich 
      --            und ergnzt den Baum entsprechend. 
      
      function CProgramm return Boolean is
         -- Mgliche Startterminale fr gegebenes Vorschauelement
         prefix : Terminals := prefTerms(Programm);
      begin
          -- leerer Baum?
         if pRtn /= null then return false; end if;
         
         -- Eingabe zu ende
         if pPrev = null then return false; end if;
         
         -- berprfen, ob Vorschau mit Eingabe bereinstimmt 
         -- => Nichtterminal nchstes Baumelement
         if prefix(TermType(pPrev.token)) then
            
            -- neuen Knoten erstellen
            pRtn := new GrKnoten'(Programm, null, null);
            pLast := pRtn;
         end if;
         
         -- Grammatikregel verwenden
         return CAnwF;
         
      end CProgramm;      
      
   begin
      -- Argument korrekt
      if pPrev = null then return null; end if;

      -- Mit Startsymbol beginnen und wenn die Terminalfolge gnzlich abgearbeitet ist,
      -- dann liegt die Tokenfolge in der Grammatik
      
      if CProgramm and then pPrev = null then return pRtn;
      else
         -- Fehlermeldung
         if error /= null then
            -- Noch Rest-Tokens
            if pPrev /= null then
               error.all := to_Unbounded_String("Parser Fehler bei Tokenposition ") &
                  to_Unbounded_String(Natural'Image(PrevPos)) &
                  to_Unbounded_String(": """) &
                  pPrev.string & to_Unbounded_String("""");
               
               -- Zeiger auf aktuelles Token mitgeben
               if pTksLeft /= null then
                  pTksLeft.all := pPrev;
               end if;
            else
               -- Tokenliste komplett abgearbeitet, aber dennoch Fehler aufgetreten
               error.all :=  to_Unbounded_String("Allgemeiner Parser Fehler: Kein vollstaendiges Programm");
               
               -- Zeiger auf aktuelles Token mitgeben
               if pTksLeft /= null then
                  pTksLeft.all := null;
               end if;
            end if;
            


               
         end if;
         return null;
      end if;
   end createGB;
   
   -------------------------------------------------------------------------------
   -- isTerminal: Knoten ist Terminal?
   
   function isTerminal (kn : GrBaum) return Boolean is
   begin
      -- Rckgabe
      return (kn.al >= TTyp'first and then kn.al <= TTyp'last);
   end isTerminal;
   
   
   -------------------------------------------------------------------------------
   -- isNonTerminal: Knoten ist Nonterminal?
   
   function isNonTerminal (kn : GrBaum) return Boolean is
   begin
      -- Rckgabe
      return (kn.al >= NTTyp'first and then kn.al <= NTTyp'last);
   end isNonTerminal;

end comtypes;

