----------------------------------------------------------------------------------
-- comtypes.ads: Paketspezifikation fr bentigte Typen und Funktion zur        --
--               Verwaltung eines Syntaxbaumes                                  --
--                                                                              --
-- Version 1.1                                                                  --
--  Feb / Mar 2008 by C. Rutetzki                                              --
---------------------------------------------------------------------------------- 

-- Token und Scanner etc. (erstellt aus aufg4.adb) 
with scannerpak; use scannerpak; 

-- Unboundes Strings
WITH ada.strings.unbounded;
USE ada.strings.unbounded;

-- Paketspezifikation fr bentigte Typen und Funktion zur Verwaltung eines 
-- Syntaxbaumes
package comtypes is
   
   
   -- Typen der Nichtterminale und Terminale
   type Alphabet is (
      -- Nichtterminale
      Programm,
      AnwF,
      Anw,
      leereAnweisung,
      Zuweisung,
      bedingteAnweisung,
      Schleife,
      Ausdruck,
      Bedingung,
      -- Terminale
      ifT,  
      nameT, 
      skipT,        
      whileT,         
      semikolonT,        
      zahlT,  
      ropT,        
      opT,         
      aufT,        
      zuT,         
      zuweisungT,        
      thenT,       
      elseT,       
      fiT,         
      doT,         
      odT,
      -- Terminale true und false wurden der Vollstndigkeit halber aufgenommen,
      -- werden aufgrund fehlender Entsprechungen in der Tokenliste aber nicht 
      -- vom Programm untersttzt
      trueT,
      falseT); 
      
      

   -- Typen der Nichtterminale
   subtype NTTyp is Alphabet range Programm .. Bedingung;
   
   -- Typen der Terminale   
   subtype TTyp is Alphabet range ifT .. falseT;
                 
   -------------------------------------------------------------------------------
   -- GrBaum: stellt eine Modellierung des Syntaxbaumes fr die 
   --             while-Grammatik dar.
   -- Felder:     - al     : legt den Typ des Knotens fest. al dient auch zur Un-
   --                        terscheidung von Terminalen und Nichtterminalen (s.o)
   --             - rest   : verweist bei jedem Knoten auf den jeweiligen rechten 
   --                        Nachbarn (Bruderknoten) sonst auf null
   --             - knoten : verweist auf den ersten Unterknoten des jeweiligen
   --                        Knotens, sofern al den Knoten als Nichtterminal 
   --                        festlegt
   --             - exp    : sofern al den Knoten als Terminal festlegt ist in
   --                        exp der Wert des Terminals enthalten (Bsp "a", ":=")
   --                        exp ist mit dem Feld string aus dem elem-Datentyp
   --                        von scannerpak identisch
   
   -- Struktur fr Knoten des Syntaxbaums der Grammatik 
   type GrKnoten;
   
   -- Zeiger auf den Baum
   type GrBaum is access GrKnoten;
   
   type GrKnoten (al : Alphabet) is record
      -- nchster Knoten (Bruder) 
      -- Bsp: bei al = ifT ist zwangslufig rest.al = Bedingung
      rest : GrBaum;
      
      -- Typ Terminal (Tokenart) oder NonTerminal?
      case al is
         -- Nichtterminal
         when Programm .. Bedingung =>
            -- Zeiger auf Unterknoten (Sohn) 
            -- Bsp: bei al = Bedingung ist zwangslufig knoten.al = Ausdruck
            knoten : GrBaum;
            
         -- Terminal  
         when ifT .. falseT =>
            -- Enthlt den "Wert" des Terminalsymbols
            -- Bsp: "2", oder ":=", etc... (entspricht dem Feld string bei Token)
            exp : unbounded_string;
         end case;
            
   end record;
   
   ------------------------------------------------------------------------------
   -- Hilfsdatentypen
   
   -- Zeiger auf String
   type PSTRING is access Unbounded_String;
   
   -- Zeiger auf Tokenliste
   type PTOKENLIST is access listOfToken;
   
   -------------------------------------------------------------------------------
   -- Erstellt aus einer Token-liste einen GrBaum (Syntaxbaum) oder im Null, wenn
   -- die Tokenliste nicht der Grammatik entspricht.
   -- (genauere Beschreibung: siehe comtypes.adb)
       
   function createGB(tks : listofToken;
                     pTksLeft : PTOKENLIST := null;
                     error : PSTRING := null) return GrBaum;

   -------------------------------------------------------------------------------
   -- Gibt an, ob der Knoten kn Terminal ist

   function isTerminal (kn : GrBaum) return Boolean;
   
   -------------------------------------------------------------------------------
   -- Gibt an, ob der Knoten kn Nichtterminal ist

   function isNonTerminal (kn : GrBaum) return Boolean;
end comtypes;

   